
/* GLUT/GL/Texture Demonstrator Program
 * Leonard D. Brown
 *
 * Acknowledgements:
 *		Portions of demo code provided by Kobus Barnard,
 *		University of Arizona
 * 
 *		ByteRaster class and methods modified from Michael Garland's
 *		LibGFX library in accordance with granted terms and conditions
 *		(see http://graphics.cs.uiuc.edu/~garland/software/libgfx.html)
 *
 * To make this code relatively independent of build paths and install
 *      paths, I've based directory paths in the VS project on environment 
 *		variables GLUT_PATH and LIBJPEG_PATH.  These environment variables
 *		must be defined by the OS should point to the root directories of 
 *		the GLUT and LIBJPEG libraries, respectively.
 *
 * If your system is lacking these APIs, you can find them here:
 *		GLUT: http://www.xmission.com/~nate/glut.html
 *	    LibJPEG: Linux: http://www.ijg.org/, 
 *				 Windows: Use the library provided on the CSc433 website OpenGL page;
 *						  other versions may cause jpeg_read_header to crash due 
 *					      to standard C library incompatibility issue with MSVC
 *						  unless you rebuild the entire library from source
 */

#include <setjmp.h>
#include <stdio.h>       
#include <stdlib.h>       

#if defined(WIN32)
#  include <windows.h>
#endif

/* RGB channel definitions */
#define RED   0
#define GREEN 1
#define BLUE  2

#define WHITE 255

#define INCREMENT 1.0

/*
// Sometimes glu.h includes glut.h--sometimes not.
*/
#ifdef MAC_OSX
#    include <OpenGL/glu.h>
#    include <GLUT/glut.h>
#else 
	#ifdef WIN32
	#	 include <GL/glu.h>
	#	 include <glut.h>
	#else 
	#    include <GL/glu.h>       
	#    include <GL/glut.h>       
	#endif
#endif

#include "raster.h"

/*
// Some debugging macros. Normally I make macros upper case; these are
// exceptions. 
*/

#define dbc(X)                                                                 \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr,                                                        \
                "dbc: "#X" is ->%c<- (0x%x, %d) on source line %d of %s\n",    \
                (int)(X), (int)(X), (int)(X), __LINE__, __FILE__);             \
    }   

#define dbp(X) \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr, "%s\n", (X));                                          \
    }   

#define dbi(X)                                                                 \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr,                                                        \
                "dbi: "#X" is %ld on source line %d of %s\n", (long)(X),       \
               __LINE__, __FILE__);                                            \
    }   

#define  dbw()                                                                 \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr,                                                        \
                "At line %d of %s\n", __LINE__, __FILE__);                     \
    }   


/* -------------------------------------------------------------------------- */

#define NOT_SET  (-99)     /* Something negative. */

/* -------------------------------------------------------------------------- */

/* 
// Some prefer to write their programs backwards, so we don't need to declare
// functions up front--I prefer to write top down, so I have added function
// declarations to the original program. 
*/

static void display_CB(void);            /* Called whenever redisplay needed */
static void idle_CB   (void);
static void key_CB(unsigned char key, int x, int y);     /* Called on key press */

/* -------------------------------------------------------------------------- */

static int debug_level = 1;  /* Used by the macros above. */


/* -------------------------------------------------------------------------- */

/* window origin on our display device, in display (pixel) coordinates */
static int x_orig = 0;
static int y_orig = 0;
/* default window size on our display device, in pixels */
static int width  = 640;
static int height = 480;

static float angle = 0.0;

/* animation flags */
static int anime_R = 0;
static int anime_G = 0; 
static int anime_B = 0;

/* mouse */
static int drag_start_x;
static int drag_start_y;
static int displacement_x;
static int displacement_y; 

/* brush color intensity */
static float color[3] = {0.4, 0.4, 0.4};

/* some hard-coded image file paths */
static char infile[]  = "pacific.jpg";
static char outfile[] = "output.jpg";

/* a raster scan of the image */
static ByteRaster *image;


/* -------------------------------------------------------------------------- */
/* IMAGE MANIPULATION EXAMPLES:
//   local helper functions for shifting the pixels
//   in a raster image vertically or horizontally
//   by one pixel; these demonstrate how to access
//   and manipulate pixel channels in the byte raster
//   object 
*/

/* vertical shift (downward) */
void vshift (ByteRaster &img, int channel, int clear)
{
	int i,j;

	for (i = img.height()-1; i >= 0; i--)
	{
		for (j = 0; j < img.width(); j++)
		{
			if (i > 0)
				/* copy previous row to current */
				img.pixel(j,i)[channel] = img.pixel(j,i-1)[channel];
			else
				/* blank the previous row to clear color */
				img.pixel(j,i)[channel] = (unsigned char) clear;
		}
	}
}

/* horizontal shift (to right) */
void hshift (ByteRaster &img, int channel, int clear)
{
	int i,j;

	for (i = img.width()-1; i >= 0; i--)
	{
		for (j = 0; j < img.height(); j++)
		{
			if (i > 0)
				/* copy previous row to current */
				img.pixel(i,j)[channel] = img.pixel(i-1,j)[channel];
			else
				/* blank the previous row to clear color */
				img.pixel(i,j)[channel] = (unsigned char) clear;
		}
	}
}

/* draw a wireframe unit cube with origin at centroid */
void draw_cube (void)
{
	glColor3d (0,1,0);

	glBegin (GL_LINES);
		/* front bottom */
		glVertex3d (-0.5,-0.5, 0.5);
		glVertex3d ( 0.5,-0.5, 0.5);
		/* front right */
		glVertex3d ( 0.5,-0.5, 0.5);
		glVertex3d ( 0.5, 0.5, 0.5);
		/* front top */
		glVertex3d ( 0.5, 0.5, 0.5);
		glVertex3d (-0.5, 0.5, 0.5);
		/* front left */
		glVertex3d (-0.5, 0.5, 0.5);
		glVertex3d (-0.5,-0.5, 0.5);
		/* top left */
		glVertex3d (-0.5, 0.5, 0.5);
		glVertex3d (-0.5, 0.5,-0.5);
		/* top right */
		glVertex3d ( 0.5, 0.5, 0.5);
		glVertex3d ( 0.5, 0.5,-0.5);
		/* bottom left */
		glVertex3d (-0.5,-0.5, 0.5);
		glVertex3d (-0.5,-0.5,-0.5);
		/* bottom right */
		glVertex3d ( 0.5,-0.5, 0.5);
		glVertex3d ( 0.5,-0.5,-0.5);
		/* back bottom */
		glVertex3d (-0.5,-0.5,-0.5);
		glVertex3d ( 0.5,-0.5,-0.5);
		/* back right */
		glVertex3d ( 0.5,-0.5,-0.5);
		glVertex3d ( 0.5, 0.5,-0.5);
		/* back top */
		glVertex3d ( 0.5, 0.5,-0.5);
		glVertex3d (-0.5, 0.5,-0.5);
		/* back left */
		glVertex3d (-0.5, 0.5,-0.5);
		glVertex3d (-0.5,-0.5,-0.5);
	glEnd ();
}

/* -------------------------------------------------------------------------- */
int main(int argc, char *argv[])
{
	/* window handle */
    int win;

	/* Read an RGB byte raster (of dimension w x h x 3) for an image */
	image = read_jpeg_image(infile);

	/* set the window dimensions to exactly fit the image. */
	if (image)
	{
	//	width  = image->width ();
		height = image->height();
	}

	/*-------------------------------------------------------------*/
	/* GLUT & GL initialization */

	/* initialize GLUT system */
    glutInit(&argc, argv);    
	/* initialize display format */
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

	/* define and get handle to the window (render) context */
    glutInitWindowSize(width, height); 
    win = glutCreateWindow("Demo"); 

	/* set window's display callback */
    glutDisplayFunc(display_CB);       
	/* background events and animation */
	glutIdleFunc (idle_CB);
	/* set window's key callback */
    glutKeyboardFunc(key_CB);

	/* set window's mouse callback */
	/*
    //glutMouseFunc(mouse_CB); 
	*/
	/* set window's mouse move with button pressed callback */
	/*
    //glutMotionFunc(mouse_move_CB);      
	*/

    /* set background clear color to black */
    glClearColor((GLclampf)0.0,(GLclampf)0.0,(GLclampf)0.0,(GLclampf)0.0);  

	glEnable     (GL_DEPTH_TEST);


	/* end GLUT & GL initialization */
	/*-------------------------------------------------------------*/

	/* setup the perspective projection */
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	/* mapping of the points on the viewplane (NDC) to window     */
	glViewport (0,0,640,480);
	/* specify the view volume, alternatively may use glFrustum() */
	/* for PERSPECTIVE projection. For ORTHOGRAPHIC projection,   */
	/* use glOrtho() given below								  */
	gluPerspective (45.0,1.333,1.0,1000.0);
	/* glOrtho (-320, 320, -240, 240, 1, 1000.0); */
	

	/* start processing events...this is the event 'pump' */
    glutMainLoop();                      

    /* Execution never reaches this point, so return value is failure */
    return EXIT_FAILURE;
}


/* -------------------------------------------------------------------------- */
/*
// Function called whenever redisplay needed (try hiding the window and then
// exposing it. The debug macro should print something.)
*/
static void display_CB(void)
{
	/* this draw callback demonstrates some common GL functionality:		*/
	/* draws 2 polygons, a polygon frame, and some pixels into the buffer	*/
	/* at arbitrary locations												*/

	/* clear the drawing buffer */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);        

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	/* specify the world-to-camera transform */
	gluLookAt (0,0,600, 0,0,0, 0,1,0);

	/* increment rotation component */
	angle += INCREMENT;

	/* Note: in addition to glRotate and glTranslate, you can construct   */
	/* transforms directly as row-major 4x4 matrices and multiply them	  */
	/* into the current active GL matrix using glMultMatrix().  The calls */
	/* to glTranslatef and glRotatef are shortcuts to do just this.  Keep */
	/* in mind that all transforms are post-multiplied onto the active GL */
	/* matrix, i.e. if the active GL matrix is C and A is your transform, */
	/* the result is always C' = CA, as opposed to C' = AC (these two	  */
	/* operations are *not* commutative in general)						  */

	// save a copy of the current model transformation state
	glPushMatrix ();
		/* notes: matrix operations are post-multiplied onto the GL matrix stack. */

		/* this polygon is 200 units in front of the image */
		glTranslatef (0,-100,200);

		/* rotate about centroid of polygon, on x axis			    */
		/* note: must center polygon since origin isn't at centroid */
		glRotatef    (angle, 1,0,0);
		glTranslatef (0,-50,0);

		/* blue polygon */
		glColor3d (0,0,1);
		glBegin (GL_QUADS);
			glVertex3d (-50,0,0);
			glVertex3d ( 50,0,0);
			glVertex3d ( 50,100,0);
			glVertex3d (-50,100,0);
		glEnd ();
	glPopMatrix ();

	glPushMatrix ();
		/* this polygon is 100 units in front of the image */
		glTranslatef (0,100,100);

		/* rotate about centroid of polygon, on y axis */
		glRotatef (-angle, 0,1,0);

		/* red polygon */
		glColor3d (1,0,0);
		glBegin (GL_QUADS);
			glVertex3d (-50,-50,0);
			glVertex3d ( 50,-50,0);
			glVertex3d ( 50, 50,0);
			glVertex3d (-50, 50,0);
		glEnd ();
	glPopMatrix ();

	/* the image, drawns as pixels */
	glPushMatrix ();

		glTranslatef (-image->width()/2.0,-image->height()/2.0,0);
		/* a simpleton way to render a raster image into a viewport. */
		int i, j;
		for (i = 0; i < image->height (); i++)
		{
			for (j = 0; j < image->width (); j++)
			{
				/* the R,G,B (byte) color of pixel(j,i) */
				glColor3d (image->pixel(j,i)[0]/255.,
						   image->pixel(j,i)[1]/255.,
						   image->pixel(j,i)[2]/255.);

				glBegin(GL_POINTS);
					/* the origin of the image is upper left corner rather */
					/* than lower right; flip to make right-side up        */
					glVertex3d (j,(image->height()-1)-i,0);
				glEnd();          
			}
		}
	glPopMatrix ();

	// draw a wirefqrame cube
	glPushMatrix ();
		glScalef (400,300,400);
		draw_cube ();
	glPopMatrix ();

	/* Complete any pending operations */
    glFlush();         

	/* Make the drawing buffer the frame buffer and vice versa */
    glutSwapBuffers(); 
}


/* -------------------------------------------------------------------------- */
/*
// This is the default background activity; this function is
// called whenever no events are detected.
*/
static void idle_CB (void)
{
	/* animation flag has been set. */
	if (anime_R)
		vshift (*image, RED, 0);
	if (anime_G)
		hshift  (*image, GREEN, 0);
	if (anime_B)
		vshift (*image, BLUE, 0);

	/* this is a crude way to regulate the pump 
	#if defined(WIN32)
		Sleep(100);
	#else
		sleep(1);
	#endif
*/

	/* state has changed: tell GLUT to redraw. */
	glutPostRedisplay();
}


/* -------------------------------------------------------------------------- */
/* Function called is called on a key press */
static void key_CB(unsigned char key, int x, int y) 
{
    dbc(key);   /* Print debugging stuff on stderr. */
    dbi(x);     /* Print debugging stuff on stderr. */
    dbi(y);     /* Print debugging stuff on stderr. */

    switch (key)
	{
		/* toggle animation flags */
		case 'r':
			anime_R = !anime_R;
			break;
		case 'g':
			anime_G = !anime_G;
			break;
		case 'b':
			anime_B = !anime_B;
			break;			
		/* save the image raster back to file */
		case 's':
			write_jpeg_image (outfile,*image);
			break;

		/* reload the image */
		case 'l':
			delete (image);
			image = read_jpeg_image (infile);
			break;
		/* quit the program -- 'hard quit' */
		case 'q':
			exit(0);
			break;
		default: 
			break;
	}
	/* state has changed: tell GLUT to redraw. */
	glutPostRedisplay ();
}

